/////////////////////////////////////////////////////////////////////////////
// GameLogic.c - Source file for game's logic
//
// Hearts Card Game
// Copyright (C) 2001-2002 Leo C. Singleton IV
// http://www.leosingleton.com/projects/hearts/
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; version 2 of the
// License. 
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Revision History
//
// 6/10/02 LCS - GameLogic.c created when MainForm.c was split into two
//		separate files (after growing to over 2100 lines); removed
//		cardsInSuit variable in IsValidPlay()
/////////////////////////////////////////////////////////////////////////////


#include "AI.h"
#include "CardDraw.h"
#include "GameLogic.h"
#include "Hearts.h"
#include "MainForm.h"


static void DealCards();
static void NewHand();
static void ResortCards(UInt8 player);
static void ComputersTurn();
static void PlayTurn();
static void TrickComplete();
static void StartNewTrick();


/////////////////////////////////////////////////////////////////////////////
// Rand()
// Generates a random number between 0 and n-1
extern UInt16 Rand(UInt16 n)
{
	return (SysRandom(0) / (1 + sysRandomMax / (n << 2))) >> 2;
}


/////////////////////////////////////////////////////////////////////////////
// IsValidPlay()
// Returns true if card is a valid play, displays message otherwise
extern Boolean IsValidPlay(UInt8 card)
{
	UInt8 i = 0;

	if (card > 51)
		return false;

	// no point cards on first trick
	if (Game.Tricks == 0)
	{
		// make sure all of the cards are not point cards; in which case,
		// we'll let them lead with anything
		while (GetSuit(Game.Cards[Game.Turn][i]) == SUIT_HEARTS ||
			Game.Cards[Game.Turn][i] == SUIT_SPADES * 13 + CARD_QUEEN)
		{
			if (i == 12)
				return true;	// all cards are points; any play is valid

			i++;
		}

		// make sure no point cards are played
		if (GetSuit(card) == SUIT_HEARTS ||
			card == (SUIT_SPADES * 13) + CARD_QUEEN)
		{
			if (Game.Turn == 0)
				ShowMessage(MsgFirstTrick);
			return false;
		}
	}

	// must lead with 2 of clubs on first trick
	if (Game.Tricks == 0)
	{
		if (Game.Cards[Game.Turn][0] == 0 && card != 0)
		{
			if (Game.Turn == 0)
				ShowMessage(MsgPlayTwoClubs);
			return false;
		}
	}

	// follow suit
	if (GetSuit(card) != Game.SuitLed && Game.SuitLed != 255)
	{
		UInt8 i;

		for (i=0; i<13; i++)
		{
			if (GetSuit(Game.Cards[Game.Turn][i]) == Game.SuitLed &&
				!Game.Used[Game.Turn][i])
			{
				if (Game.Turn == 0)
					ShowMessage(MsgFollowSuit);
				return false;
			}
		}
	}

	// hearts not broken
	if (GetSuit(card) == SUIT_HEARTS && !Game.HeartsBroken)
	{
		if (Game.Play[Game.Turn == 0 ? 3 : Game.Turn - 1] == 255)
		{
			// this was the first play of the trick
			UInt8 i;
			
			for (i=0; i<13; i++)
			{
				if (GetSuit(Game.Cards[Game.Turn][i]) != SUIT_HEARTS &&
					!Game.Used[Game.Turn][i])
				{
					if (Game.Turn == 0)
						ShowMessage(MsgHeartsNotBroken);
					return false;
				}
			}
		}
	}

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// GetSuit()
// Helper function for AI routines to determine card suit
extern UInt8 GetSuit(UInt8 card)
{
	return card / 13;
}


/////////////////////////////////////////////////////////////////////////////
// GetCard()
// Helper function for AI routines to determine card value
extern UInt8 GetCard(UInt8 card)
{
	return card % 13;
}


/////////////////////////////////////////////////////////////////////////////
// DealCards()
// Initializes card arrays by random dealing
static void DealCards()
{
	UInt8 cardAssignments[52];
	UInt8 cardCount[4];
	UInt8 player, card, turn;
	
	// initialize card assignment array
	for (card=0; card<52; card++)
		cardAssignments[card] = 255;
	
	// assign cards to players (let players take turns choosing a random card
	// from the remaining cards)
	for (turn=0; turn<13; turn++)
	{
		for (player=0; player<4; player++)
		{
			// choose a random card from the deck
			card = Rand(52);
			
			// if the chosen card has already been taken, try the next one in
			// the deck
			while (cardAssignments[card] != 255)
			{
				card++;
				if (card >= 52)
					card = 0;
			}
			
			// assign the chosen card to the current player
			cardAssignments[card] = player;
		}
	}
	
	// initialize the card count array
	for (player=0; player<4; player++)
		cardCount[player] = 0;
		
	// distribute the cards to the players (meanwhile, mark the cards as
	// unused)
	for (card=0; card<52; card++)
	{
		player = cardAssignments[card];
		Game.Cards[player][cardCount[player]] = card;
		Game.Used[player][cardCount[player]++] = false;
	}
}


/////////////////////////////////////////////////////////////////////////////
// NewGame()
// Resets all variables to begin a new game
extern void NewGame()
{
	UInt8 i;

	DeallocateMainScreenBuffers();

	Game.PassTo = 0;
	for (i=0; i<4; i++)
		Game.TotalScore[i] = 0;

	NewHand();
}


/////////////////////////////////////////////////////////////////////////////
// NewHand()
// Resets some variables to begin a new hand within the same game
static void NewHand()
{
	UInt8 i;

	Game.HeartsBroken = false;

	for (i=0; i<4; i++)
	{
		Game.RoundScore[i] = 0;
		Game.Play[i] = 255;
	}

	for (i=0; i<13; i++)
		Game.Invert[i] = false;
	
	DealCards();

	if (Game.PassTo != 3)
	{
		// this is a hand where we pass
		Game.GameStatus = GAMESTATUS_SELECTPASS;

		// clear cards selected to pass
		Game.CardPassCount = 0;

		PaintScreen(true);
	}
	else
	{
		// we won't pass in this hand - skip to playing
		Game.PassTo = 0;
		StartPlay();
	}
}


/////////////////////////////////////////////////////////////////////////////
// SwapCards()
// Swaps 3 cards between player1 and player2, determined by the cards passed
// by the arrays
extern void SwapCards(UInt8 player1, UInt8 *pCards1, UInt8 player2,
					  UInt8 *pCards2)
{
	UInt8 card1, card2, temp, i;

	for (i=0; i<3; i++)
	{
		card1 = pCards1[i];
		card2 = pCards2[i];
		temp = Game.Cards[player1][card1];
		Game.Cards[player1][card1] = Game.Cards[player2][card2];
		Game.Cards[player2][card2] = temp;
	}

	ResortCards(player1);
	ResortCards(player2);
}


/////////////////////////////////////////////////////////////////////////////
// ResortCards()
// Resorts cards after a pass has occurred
static void ResortCards(UInt8 player)
{
	UInt8 x, y, temp;

	for(x=0; x<12; x++)
	{
		for(y=0; y<12; y++)
		{
			if (Game.Cards[player][y] > Game.Cards[player][y+1])
			{
				// swap places
				temp = Game.Cards[player][y];
				Game.Cards[player][y] = Game.Cards[player][y+1];
				Game.Cards[player][y+1] = temp;
			}
		}
	}
}


/////////////////////////////////////////////////////////////////////////////
// StartPlay()
// Called after pass completed to begin playing a hand
extern void StartPlay()
{
	UInt8 i;
	
	Game.Tricks = 0;
	Game.AnimCount = 0;
	for (i=0; i<13; i++)
		Game.Invert[i] = false;
	for (i=0; i<4; i++)
		Game.Play[i] = 255;

	PaintScreen(true);
	
	// quick fix to draw screen on the fourth hand
	if (Game.PassTo == 0)
	{
		DrawPlayerCardsAll();
		DrawPlayerNames();
	}

	// find who has the 2 of clubs
	for (i=0; i<4; i++)
		if (Game.Cards[i][0] == 0)
			Game.Turn = i;

	StartNewTrick();
}


/////////////////////////////////////////////////////////////////////////////
// ComputersTurn()
// Called to play a computer player's turn
static void ComputersTurn()
{
	UInt8 card;

	// determine move
	card = AISelectCard(Game.Turn);

	// set card as used
	Game.Used[Game.Turn][card] = true;
	Game.Play[Game.Turn] = card;

	// set suit led if this was the first play
	if (Game.SuitLed == 255)
		Game.SuitLed = GetSuit(Game.Cards[Game.Turn][card]);

	// update HeartsBroken
	if (GetSuit(Game.Cards[Game.Turn][card]) == SUIT_HEARTS)
		Game.HeartsBroken = true;

	// start animation to middle
	switch (Game.Turn)
	{
	case 1:
		Game.AnimStartX = PLAYER1CARD_XI;
		Game.AnimStartY = PLAYER1CARD_Y;
		Game.AnimEndX = PLAYER1CARD_X;
		Game.AnimEndY = PLAYER1CARD_Y;
		break;

	case 2:
		Game.AnimStartX = PLAYER2CARD_X;
		Game.AnimStartY = PLAYER2CARD_YI;
		Game.AnimEndX = PLAYER2CARD_X;
		Game.AnimEndY = PLAYER2CARD_Y;
		break;

	case 3:
		Game.AnimStartX = PLAYER3CARD_XI;
		Game.AnimStartY = PLAYER3CARD_Y;
		Game.AnimEndX = PLAYER3CARD_X;
		Game.AnimEndY = PLAYER3CARD_Y;
		break;
	}
	Game.AnimLastX = Game.AnimStartX;
	Game.AnimLastY = Game.AnimStartY;
	Game.AnimProgress = 0;
	Game.AnimCard = GetCard(Game.Cards[Game.Turn][card]);
	Game.AnimSuit = GetSuit(Game.Cards[Game.Turn][card]);
	Game.GameStatus = GAMESTATUS_ANIMATE;

	// prepare for next turn
	Game.Turn++;
	if (Game.Turn > 3)
		Game.Turn = 0;
}


/////////////////////////////////////////////////////////////////////////////
// PlayTurn()
// Called to begin a new trick
static void PlayTurn()
{
	if (Game.Turn == 0)
	{
		// human player goes first
		ShowMessage(MsgSelectCardPlay);
		Game.GameStatus = GAMESTATUS_SELECTCARD;
	}
	else
	{
		// computer player's turn
		ShowMessage(MsgBlank);
		ComputersTurn();
	}
}


/////////////////////////////////////////////////////////////////////////////
// PostAnimate()
// Called when an animation routine is completed
extern void PostAnimate()
{
	if (Game.Play[Game.Turn] == 255)
		PlayTurn();
	else
		TrickComplete();
}


/////////////////////////////////////////////////////////////////////////////
// TrickComplete()
// Called after all 4 players have played
//   (note: this routine is called 5 times - once for each fly-away animation
//    and once to move on to the next trick)
static void TrickComplete()
{
	UInt8 suit, score = 0;
	Int16 highCard = -1, highPlayer = -1;
	UInt8 i;
	UInt8 playerSuit, playerCard;
	RectangleType rect;

#ifdef COLOR
	IndexedColorType oldColor;
#endif // COLOR

	if (Game.AnimCount == 0)
	{
		// this is the first time TrickComplete() was called
		suit = GetSuit(Game.Cards[Game.Turn][Game.Play[Game.Turn]]);

		for (i=0; i<4; i++)
		{
			// get card and suit of player i's card
			playerSuit = GetSuit(Game.Cards[i][Game.Play[i]]);
			playerCard = GetCard(Game.Cards[i][Game.Play[i]]);

			// determine winner of trick
			if (playerSuit == suit && playerCard > highCard)
			{
				highCard = playerCard;;
				highPlayer = i;
			}

			// update scores
			if (playerSuit == SUIT_HEARTS)
				score++;

			if (playerSuit == SUIT_SPADES && playerCard == CARD_QUEEN)
				score += 13;
		}

		Game.TrickWinner = highPlayer;

		// add score to winner of trick
		Game.RoundScore[highPlayer] += score;
	}

	if (Game.AnimCount == 4)
	{
		// All post-trick animations are complete
		Game.Tricks++;

		if (Game.Tricks == 13)
		{
			// game over
			UInt8 shootTheMoon = 255;
			Boolean bGameOver = false;
			
			// check for shooting the moon
			for (i=0; i<4; i++)
			{
				if (Game.RoundScore[i] == 26)
					shootTheMoon = i;
			}
			
			if (shootTheMoon != 255)
			{
				// special scoring for shooting the moon
				for (i=0; i<4; i++)
				{
					if (i == shootTheMoon)
						Game.RoundScore[i] = 0;
					else
						Game.RoundScore[i] = 26;
				}
			}

			// total scores
			for (i=0; i<4; i++)
				Game.TotalScore[i] += Game.RoundScore[i];

			// check for game over
			for (i=0; i<4; i++)
			{
				if (Game.TotalScore[i] >= 100)
					bGameOver = true;
			}

			// show score dialog
			DoScoreDialog();

			if (!bGameOver)
			{
				// start a new hand
				NewHand();
				return;
			}

			// game over
			Game.GameStatus = GAMESTATUS_NOGAME;
			PaintScreen(true);
			return;
		}

		// winner of the last trick leads the next one
		Game.Turn = Game.TrickWinner;
		StartNewTrick();
		return;
	}

	// this isn't a special case, so just play the animation
	switch (Game.AnimCount)
	{
	case 0:
		Game.AnimStartX = PLAYER0CARD_X;
		Game.AnimStartY = PLAYER0CARD_Y;
		break;

	case 1:
		Game.AnimStartX = PLAYER1CARD_X;
		Game.AnimStartY = PLAYER1CARD_Y;
		break;

	case 2:
		Game.AnimStartX = PLAYER2CARD_X;
		Game.AnimStartY = PLAYER2CARD_Y;
		break;

	case 3:
		Game.AnimStartX = PLAYER3CARD_X;
		Game.AnimStartY = PLAYER3CARD_Y;
		break;
	}
	switch (Game.TrickWinner)
	{
	case 0:
		Game.AnimEndX = PLAYER0CARD_X;
		Game.AnimEndY = PLAYER0CARD_YI;
		break;

	case 1:
		Game.AnimEndX = PLAYER1CARD_XI;
		Game.AnimEndY = PLAYER1CARD_Y;
		break;

	case 2:
		Game.AnimEndX = PLAYER2CARD_X;
		Game.AnimEndY = PLAYER2CARD_YI;
		break;

	case 3:
		Game.AnimEndX = PLAYER3CARD_XI;
		Game.AnimEndY = PLAYER3CARD_Y;
		break;
	}
	Game.AnimLastX = Game.AnimStartX;
	Game.AnimLastY = Game.AnimStartY;
	Game.AnimProgress = 0;
	Game.AnimCard = GetCard(Game.Cards[Game.AnimCount][Game.Play[Game.AnimCount]]);
	Game.AnimSuit = GetSuit(Game.Cards[Game.AnimCount][Game.Play[Game.AnimCount]]);
	Game.GameStatus = GAMESTATUS_ANIMATE;
	
	// this is a simple workaround to prevent the card from appearing to stay
	// on the screen after it has flown off
	//   (Note: This code causes the flickering to occur when the cards fly
	//    away.  To make the animation truly go away, it needs to be fixed in
	//    the buffering of the Animation function)
	rect.topLeft.x = Game.AnimStartX;
	rect.topLeft.y = Game.AnimStartY;
	rect.extent.x = 20;
	rect.extent.y = 30;
#ifdef COLOR
	oldColor = WinSetBackColor(BACKGROUND_COLOR);
#endif // COLOR
	WinEraseRectangle(&rect, 0);
#ifdef COLOR
	WinSetBackColor(oldColor);
#endif // COLOR

	// increment the animation counter (this will make the next card fly off
	// next)
	Game.AnimCount++;
}


/////////////////////////////////////////////////////////////////////////////
// StartNewTrick()
// Called to begin a new trick
static void StartNewTrick()
{
	UInt8 i;

	// reset variables
	Game.SuitLed = 255;
	Game.AnimCount = 0;
	for (i=0; i<4; i++)
		Game.Play[i] = 255;

	PlayTurn();
}


/////////////////////////////////////////////////////////////////////////////
// EndGame()
// Returns to the intro screen
extern void EndGame()
{
	FormType *form;
	UInt8 i;

	form = FrmGetActiveForm();
	FrmHideObject(form, FrmGetObjectIndex(form, MainPassButton));
	FrmDrawForm(form);

	Game.CardPassCount = 0;
	for (i=0; i<13; i++)
		Game.Invert[i] = false;
	
	Game.GameStatus = GAMESTATUS_NOGAME;
	PaintScreen(false);
}


