/////////////////////////////////////////////////////////////////////////////
// Hearts.c - Source file for main application
//
// Hearts Card Game
// Copyright (C) 2001 Leo C. Singleton IV
// http://www.leosingleton.com/projects/hearts/
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; version 2 of the
// License. 
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Revision History
//
// 8/8/01 LCS - changed Ptr to MemPtr; removed #defines needed to upgrade
//		from Palm OS SDK 3.1 to SDK 4.0
// 8/9/01 LCS - updated RomVersionCompatible() to check for color support on
//		color builds of Hearts
// 8/21/01 LCS - RomVersionCompatible() now checks for a normal launch flag
//		before doing any other compatibility tests.  This should prevent
//		hard resets on Visor Prisms.  Changed winScreenModeGet to
//		winScreenModeSet in StopApplication().  Added gResetDisplay.
// 9/5/01 LCS - moved RomVersionCompatible() call to the
//		sysAppLaunchCmdNormalLaunch block of PilotMain().  Maybe this will
//		fix the crashing on Visor Prisms.
// 9/28/01 LCS - removed debugging and dead code
// 12/30/01 LCS - added wide character support
// 6/6/02 LCS - added support for new v1.3 preferences
// 6/10/02 LCS - initialized err to 0 in PilotMain(); added
//		#include "CardDraw.h"
/////////////////////////////////////////////////////////////////////////////


#include "CardDraw.h"
#include "Hearts.h"
#include "MainForm.h"


// global variables
UInt8 AnimCount=0;

Prefs_Type Prefs;
Game_Type Game;

// Local function declarations
static void EventLoop();
static Err StartApplication();
static void StopApplication();
static Boolean ApplicationHandleEvent(EventPtr event);
static void InitGameStruct();
static Err RomVersionCompatible (UInt32 requiredVersion, UInt16 launchFlags);

#define version20						0x02000000


#ifdef COLOR
UInt32 gOldDepth;
Boolean gResetDisplay;
#endif


Boolean WideCharacterSupport;


/////////////////////////////////////////////////////////////////////////////
// PilotMain()
// Application Entry Point
UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err = 0;

	switch (launchCode)
	{
	case sysAppLaunchCmdNormalLaunch:
		err = RomVersionCompatible(version20, launchFlags);
		if (err)
			return err;

		if ((err = StartApplication()) == 0)
		{
			EventLoop();
			StopApplication();
		}
		break;

	default:
		break;
	}

	return err;
}


/////////////////////////////////////////////////////////////////////////////
// StartApplication()
// Handles application startup
// (subroutine of PilotMain())
static Err StartApplication()
{
	UInt16 prefSize;
	UInt8 i;

	UInt32 value;
	Err error;

	Game.CurrentSelection = 255;

	// determine if we need wide character support
	error = FtrGet(sysFtrCreator, sysFtrNumIntlMgr, &value);
	if ((value != 0) && (error == 0))
		WideCharacterSupport = true;
	else
		WideCharacterSupport = false;

	// load data from preferences
	prefSize = sizeof(Prefs_Type);
	if (PrefGetAppPreferences(appCreatorID, appPrefID, &Prefs, &prefSize,
		true) == noPreferenceFound)
	{
		// load default preferences
		Prefs.AnimationDelay = PREF_ANIMSMOOTH_MED;
		Prefs.AnimationSpeed = Prefs.AnimationDelay * PREF_ANIMSPEED_MED;
		
		for (i=0; i<4; i++)
		{
			MemHandle strH;
			char *pStr;

			// get a pointer to the string data
			strH = DmGetResource(strRsc, PlayerName1 + i);
			if (strH)
			{
				pStr = (char*)MemHandleLock(strH);
				StrCopy(Prefs.Names[i], pStr);
				MemHandleUnlock(strH);
			}
		}

		Prefs.MarkPlayableCards = true;
		Prefs.Difficulty = PREF_DIFFICULTY_MEDIUM;
	}

	prefSize = sizeof(Game_Type);
	if (PrefGetAppPreferences(appCreatorID, appGameID, &Game, &prefSize,
		false) == noPreferenceFound)
		InitGameStruct();


	// initialize drawing routines
	PreDraw();
	
	FrmGotoForm(MainForm);

	return 0;
}


/////////////////////////////////////////////////////////////////////////////
// StopApplication()
// Handles application shutdown and stores settings
// (subroutine of PilotMain())
static void StopApplication()
{
	// save preferences
	PrefSetAppPreferences(appCreatorID, appPrefID, appPrefVersion, &Prefs,
		sizeof(Prefs), true);
	
	PrefSetAppPreferences(appCreatorID, appGameID, appPrefVersion, &Game,
		sizeof(Game), false);

	// close drawing routines
	PostDraw();

#ifdef COLOR
	if (gResetDisplay)
		WinScreenMode(winScreenModeSet, NULL, NULL, &gOldDepth, NULL);
#endif // COLOR
}


/////////////////////////////////////////////////////////////////////////////
// EventLoop()
// The One and Only Event Loop
// (subroutine of PilotMain())
static void EventLoop()
{
	EventType event;
	UInt16 error;

	do
	{
		EvtGetEvent(&event, Prefs.AnimationDelay);
		if (!SysHandleEvent(&event))
			if (!MenuHandleEvent(0, &event, &error))
				if (!ApplicationHandleEvent(&event))
					FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}


/////////////////////////////////////////////////////////////////////////////
// ApplicationHandleEvent()
// Initializes the proper form event handler
static Boolean ApplicationHandleEvent(EventPtr event)
{
	FormPtr form;
	UInt16 formID;
	Boolean handled = false;

	if (event->eType == frmLoadEvent)
	{
		formID = event->data.frmLoad.formID;
		form = FrmInitForm(formID);
		FrmSetActiveForm(form);

		switch (formID)
		{
		case MainForm:
			FrmSetEventHandler(form, MainFormHandleEvent);
			break;
		}

		handled = true;
	}

	return handled;
}


/////////////////////////////////////////////////////////////////////////////
// InitGameStruct()
// Initializes the game
static void InitGameStruct()
{
	UInt8 i;
	
	Game.CardPassCount = 0;
	for (i=0; i<13; i++)
		Game.Invert[i] = false;
	
	Game.GameStatus = GAMESTATUS_NOGAME;

	Game.PassTo = 0;
}


/////////////////////////////////////////////////////////////////////////////
// RomVersionCombatible()
// from CodeWarrior template
static Err RomVersionCompatible (UInt32 requiredVersion, UInt16 launchFlags)
{
	UInt32 romVersion;
#ifdef COLOR
	UInt32 depthP;
#endif

	// See if we have at least the minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion)
	{
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
			(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
		{
			FrmAlert (RomIncompatibleAlert);
		
			// Pilot 1.0 will continuously relaunch this app unless we switch to 
			// another safe one.
			if (romVersion < 0x02000000)
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
		}
		
		return sysErrRomIncompatible;
	}

#ifdef COLOR
	if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
		(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
	{
		if (romVersion < 0x03503000)
		{
			FrmAlert(ColorNotSupportedAlert);
			return sysErrRomIncompatible;
		}
	
		WinScreenMode(winScreenModeGetSupportedDepths, NULL, NULL, &depthP,
			NULL);

		if (!(0x80 & depthP))
			return sysErrRomIncompatible;

		WinScreenMode(winScreenModeGet, NULL, NULL, &gOldDepth, NULL);
		depthP = 8;
		WinScreenMode(winScreenModeSet, NULL, NULL, &depthP, NULL);
		gResetDisplay = true;
	}
	else
		gResetDisplay = false;
#endif // COLOR

	return 0;
}

